/*-------------------------------------------------------------------------
 * Copyright (c) 2000-2002 Kenneth W. Sodemann (stuffle@charter.net)
 *-------------------------------------------------------------------------
 * sendmsg
 *
 * Synopsis:
 *   Procedures used to send out a predefined e-mail message.
 *
 * $Id: sendmsg.c,v 1.4 2002/01/30 00:40:02 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/wait.h>

#include "sendmsg.h"

#ifndef FALSE
#   define FALSE      0
#endif
#ifndef TRUE
#   define TRUE       1
#endif
#ifndef READ_PIPE
#   define READ_PIPE  0
#endif
#ifndef WRITE_PIPE
#   define WRITE_PIPE 1
#endif

#define INVALID_PR_NUM  -1

#define NEW_PR_MSG       0
#define ASSIGN_MSG       1
#define UPDATE_MSG       2

/*
 * Define the template files
 */
#define NEW_PR_TEMPLATE "new_pr.msg"
#define ASSIGN_TEMPLATE "assign.msg"
#define UPDATE_TEMPLATE "update.msg"

/*
 * Define how we should wrap the messages
 */
#define MAX_LINE_LEN   75
#define BREAK_ZONE_LEN 10

/*
 * Send out this message if the template file cannot be opened.
 */
#define ERR_MSG "Subject:  PRepS msg template missing.\n\nPRepS is supposed to be sending you a message about a PR, but the\ntemplate file is missing or otherwise not openable.  Contact your\nPRepS administrator about this..."

/*
 * Define the valid tags.
 */
#define TAG_START       '<'
#define TAG_END         '>'
#define TITLE_TAG       "TITLE"
#define DESCR_TAG       "DESCR"
#define FIX_TAG         "FIX"
#define STAT_TAG        "STAT"
#define SEVR_TAG        "SEVR"
#define RESP_TAG        "RESP"
#define SUBMIT_TAG      "SUBMITTER"
#define PROJECT_TAG     "PRJ"
#define PR_NUM_TAG      "NUM" 
#define PR_TYPE_TAG     "PR_TYPE"
#define MAX_TAG_LEN     9 
#define MAX_BUFF_LEN    255

typedef enum
{
   PR_NUM,
   PROJECT,
   PR_TITLE,
   DESCRIPTION,
   FIX_DESCRIPTION,
   STATUS,
   SEVERITY,
   PROBLEM_TYPE,
   RESPONSIBLE,
   SUBMITTER,
   INVALID_TAG
} TagType;

typedef enum
{
   NORMAL,
   IN_A_TAG
} StateType;

static TagType get_tag_type( const char *tag )
{
   char tmp_buf[MAX_TAG_LEN + 1];
   int  len;
   int  i;

   len = strlen( tag );
   assert( len <= MAX_TAG_LEN );

   /*
    * Brute force method.  Upcase the tag, and compare against each 
    * one of the defined tags.  Return upon match.  Return INVALID_TAG
    * if tag does not match any.
    */
   for( i = 0; i < len; i++ )
      {
      tmp_buf[i] = toupper( tag[i] );
      }
   tmp_buf[len] = '\000';

   if( strcmp( tmp_buf, PR_NUM_TAG ) == 0 )
      {
      return PR_NUM;
      }

   if( strcmp( tmp_buf, PROJECT_TAG ) == 0 )
      {
      return PROJECT;
      }

   if( strcmp( tmp_buf, TITLE_TAG ) == 0 )
      {
      return PR_TITLE;
      }

   if( strcmp( tmp_buf, DESCR_TAG ) == 0 )
      {
      return DESCRIPTION;
      }

   if( strcmp( tmp_buf, FIX_TAG ) == 0 )
      {
      return FIX_DESCRIPTION;
      }

   if( strcmp( tmp_buf, STAT_TAG ) == 0 )
      {
      return STATUS;
      }

   if( strcmp( tmp_buf, SEVR_TAG ) == 0 )
      {
      return SEVERITY;
      }

   if( strcmp( tmp_buf, PR_TYPE_TAG ) == 0 )
      {
      return PROBLEM_TYPE;
      }

   if( strcmp( tmp_buf, RESP_TAG ) == 0 )
      {
      return RESPONSIBLE;
      }

   if( strcmp( tmp_buf, SUBMIT_TAG ) == 0 )
      {
      return SUBMITTER;
      }

   return INVALID_TAG;
}


/*-------------------------------------------------------------------------
 * wrap_write
 *
 * Write out the buffer, wrapping lines at with a window of BREAK_ZONE_LEN
 * characters from MAX_LINE_LEN.
 *
 * This function is dirt simple.  It goes through the buffer, writing
 * line by line.  If a line gets to be MAX_LINE_LEN characters long, 
 * this routine will look in a break zone for the last white space 
 * character and break the line there.  If a white space charcter
 * cannot be found within the break zone, a hyphen will be placed in
 * the middle of the break zone, and the line will be broken at that
 * point.
 *
 * This is a dirt simple and brain-dead way of doing this, but it should
 * work for our purposes.
 *-------------------------------------------------------------------------
 */
static void
wrap_write (int fd, const char *buf, size_t cnt)
{
   char buffer[MAX_LINE_LEN + 1];
   int  i;          /* index into full buf */
   int  buff_len;   /* index into master buff */ 
   int  found_break;

   /*
    * If the line isn't long enough to ever wrap, just write it
    * out...
    */
   if (cnt < MAX_LINE_LEN)
      {
      write (fd, buf, cnt);
      return;
      }

   buff_len = 0;
   i = 0;
   while (i < cnt) 
      {
      buffer[buff_len++] = buf[i++];

      if (buf[i - 1] == '\n')
         {
         /*
          * The line broke on its own.  Dump it out.
          */
         write (fd, buffer, buff_len);
         buff_len = 0;
         }
      else if (buff_len == MAX_LINE_LEN)
         {
         /*
          * Look for the last white space char within the break 
          * zone (that would be the FIRST one if you are going
          * backwards, which we are).
          */
         found_break = FALSE;
         while (!found_break &&
                buff_len > MAX_LINE_LEN - BREAK_ZONE_LEN)
            {
            if (isspace (buffer[--buff_len]))
               {
               buffer[buff_len++] = '\n';
               found_break = TRUE;
               }
            else
               {
               --i;
               }
            }
         if (!found_break)
            {
            /* 
             * The last word must be LONG.  Hyphenate it.
             */
            i += BREAK_ZONE_LEN / 2;
            buff_len += (BREAK_ZONE_LEN / 2);
            buffer[buff_len++] = '-';
            buffer[buff_len++] = '\n';
            }
         write (fd, buffer, buff_len);
         buff_len = 0;
         }
      }

   /* 
    * If the end of the line is not a '\n' character, we
    * did not write it out yet.
    */
   if (buffer[buff_len - 1] != '\n')
      {
      write (fd, buffer, buff_len);
      }
}

/*-------------------------------------------------------------------------
 * write_msg
 *
 * Using the defined template, write the e-mail message replacing the
 * embedded tags with the proper information.
 *
 * Algorithm:
 *
 * open template file
 * loop
 *   read char from template file
 *   switch (state)
 *     case NORMAL:
 *       if char is tag start
 *         set state to IN_A_TAG
 *         set tag char counter to zero
 *       else
 *         write char to msg file
 *     case IN_A_TAG:
 *       if char is tag start
 *         write raw text read to msg file
 *         set tag char counter to zero
 *       elseif char is tag end
 *         if tag text matches one of the tags
 *           write associated text to msg file
 *         else
 *           write raw text read to msg file
 *         set state to NORMAL
 *       else
 *         if tag char counter + 1 > MAX_TAG_LEN
 *           write raw text read to msg file
 *           set state to NORMAL
 *         else
 *           save char in current tag buffer
 *     default:
 *       assert( FALSE )
 *   end switch
 * end loop
 * close template file
 *
 *-------------------------------------------------------------------------
 */
static void
write_msg( int         fd,
           int         msg_type,
           int4        pr_num,
           const text *project,
           const text *pr_title,
           const text *descr,
           const text *fix_descr,
           const text *status,
           const text *severity,
           const text *pr_type,
           const text *responsible,
           const text *submitter )
{
   StateType  state = NORMAL;
   char       curr_char;
   char       buffer[MAX_BUFF_LEN + 1]; 
   int        buffer_len = 0;
   FILE      *in_file = NULL;
   TagType    curr_tag_type;
   text      *curr_text = NULL;

   /*
    * Build the file name for the template file, and open it.
    */
   strcpy( buffer, DATA_DIR );
   strcat( buffer, "/" );
   switch( msg_type )
      {
      case NEW_PR_MSG:
         strcat( buffer, NEW_PR_TEMPLATE );
         break;
      case ASSIGN_MSG:
         strcat( buffer, ASSIGN_TEMPLATE );
         break;
      case UPDATE_MSG:
         strcat( buffer, UPDATE_TEMPLATE );
         break;
      default:
         assert( FALSE );
      };
   /*
    * If we cannot open the template file, bug the recipients with an
    * error message e-mail.  Eventually someone will tell the PRepS
    * administrator that they need to fix something.
    */
   in_file = fopen( buffer, "r" );
   if( !in_file )
      {
      write( fd, ERR_MSG, strlen( ERR_MSG ) );
      return;
      }

   /*
    * Loop through the file and implement our state machine as defined
    * in the function description.  Buffer the output for efficiency
    * (it also make the state machine a little simpler).
    */
   while( (curr_char = fgetc( in_file )) != EOF )
      {
      switch( state )
         {
         case NORMAL:
            if( curr_char == TAG_START )
               {
               state = IN_A_TAG;
               /* 
                * We are sharing the tag buffer with the regular
                * output buffer.  Flush the output buffer before
                * we start building the tag.
                */
               write( fd, buffer, buffer_len );
               buffer[0] = TAG_START;
               buffer_len = 1;
               }
            else
               {
               buffer[buffer_len] = curr_char;
               buffer_len++;
               if( buffer_len == MAX_BUFF_LEN )
                  {
                  write( fd, buffer, buffer_len );
                  buffer_len = 0;
                  }
               }
            break;

         case IN_A_TAG:
            if( curr_char == TAG_START )
               {
               write( fd, buffer, buffer_len );
               buffer[0] = TAG_START;
               buffer_len = 1;
               }
            else if( curr_char == TAG_END )
               {
               /*
                * When calling get_tag_type, skip the starting
                * TAG_START character that we have in the first
                * position in the buffer.
                */
               buffer[buffer_len] = '\000';
               curr_tag_type = get_tag_type( &buffer[1] );

               /*
                * Output for non-text types will be handled in the
                * switch.  For text types, the switch sets curr_text,
                * and the output to the file is handled after the
                * switch.
                *
                * Currently PR_NUM is the only special case (non-text
                * output).
                */
               switch( curr_tag_type )
                  {
                  case PR_NUM:
                     if( pr_num == INVALID_PR_NUM )
                        {
                        buffer[buffer_len] = TAG_END;
                        buffer_len++;
                        }
                     else
                        {
                        sprintf( buffer, "%d", pr_num );
                        buffer_len = strlen( buffer );
                        }
                     write( fd, buffer, buffer_len );
                     break;

                  case PROJECT:
                     curr_text = (text *)project;
                     break;

                  case PR_TITLE:
                     curr_text = (text *)pr_title;
                     break;

                  case DESCRIPTION:
                     curr_text = (text *)descr;
                     break;

                  case FIX_DESCRIPTION:
                     curr_text = (text *)fix_descr;
                     break;

                  case STATUS:
                     curr_text = (text *)status;
                     break;

                  case SEVERITY:
                     curr_text = (text *)severity;
                     break;

                  case PROBLEM_TYPE:
                     curr_text = (text *)pr_type;
                     break;

                  case RESPONSIBLE:
                     curr_text = (text *)responsible;
                     break;

                  case SUBMITTER:
                     curr_text = (text *)submitter;
                     break;

                  case INVALID_TAG:
                     curr_text = NULL;
                     break;

                  default:
                     assert( FALSE );
                  } /* end tag type switch */

               /*
                * If the current tag is a text type (or INVALID_TAG), 
                * dump the corresponding text to the file.  
                *
                * If curr_text is NULL, it indicates that the tag was 
                * either invalid, or was not supposed to be used with 
                * the current type of message.  In that case, just dump 
                * the tag.
                */
               if( curr_tag_type != PR_NUM )
                  {
                  if( curr_text )
                     {
                     wrap_write( fd, VARDATA(curr_text),
                                 VARSIZE(curr_text) - VARHDRSZ );
                     }
                  else
                     {
                     buffer[buffer_len] = TAG_END;
                     buffer_len++;
                     write( fd, buffer, buffer_len );
                     }
                  }

               buffer_len = 0;
               state = NORMAL;
               }
            else
               {
               buffer[buffer_len] = curr_char;
               buffer_len++;
               /*
                * If we have gone past the maximum possible length for a
                * buffer (remember the starting TAG_START char), there is
                * no way that we have a valid tag.  Put us back into
                * NORMAL state.
                */ 
               if( buffer_len > MAX_TAG_LEN + 1 )
                  {
                  state = NORMAL;
                  }
                   
               }
            break;

         default:
            assert( FALSE );
         } /* end state switch */
      } /* end while */

   /*
    * Flush whatever is left in the buffer.
    */
   write( fd, buffer, buffer_len );
   fclose( in_file );

   return;
}


/*-------------------------------------------------------------------------
 * sendmsg
 *
 * fork sendmail, and feed it the message by calling write_msg with the
 * write end of the pipe as the fd.
 *
 * If any message type should not use one of the text replacements,
 * sendmsg should be called with a NULL for that argument 
 * (or INVALID_PR_NUM if pr_num is not to be used with the msg_type).
 *-------------------------------------------------------------------------
 */
static void
sendmsg( int         msg_type,
         const text *email_list,
         int4        pr_num,
         const text *project,
         const text *pr_title,
         const text *descr,
         const text *fix_descr,
         const text *status,
         const text *severity,
         const text *pr_type,
         const text *responsible,
         const text *submitter )
{
   int    fork_result;
   int    pipe_files[2];
   int    fd;
   int    child_status;

   /*
    * Build the proper message.  Get the name of the file that contains
    * that message.
    */
   if( pipe( pipe_files ) != -1 )
      {
      fork_result = fork();
      switch( fork_result )
         {
         case -1:
            syslog( LOG_ERR, "fork failed: %m" );
            close( pipe_files[READ_PIPE] );
            close( pipe_files[WRITE_PIPE] );
            _exit( -1 );

         case 0:
            {
            char *email_list_chr = NULL;
            /*
             * We are in the child.  Prepare the child for executing
             * sendmail by duping stdin into the READ end of the pipe.
             * The parent will then send the mail message through the
             * pipe.
             */
            if( close( 0 ) == -1 )
               {
               syslog( LOG_ERR, "failed to close stdin: %m" );
               close( pipe_files[READ_PIPE] );
               close( pipe_files[WRITE_PIPE] );
               _exit( -1 );
               }

            fd = dup( pipe_files[READ_PIPE] );
            if( fd != 0 )
               {
               syslog( LOG_ERR, "failed to dup read pipe: %m" );
               close( pipe_files[READ_PIPE] );
               close( pipe_files[WRITE_PIPE] );
               _exit( -1 );
               } 

            close( pipe_files[READ_PIPE] );
            close( pipe_files[WRITE_PIPE] );

            email_list_chr = 
               (char *)malloc( VARSIZE(email_list) + 1 - VARHDRSZ );
            if (email_list_chr)
               {
               strncpy( email_list_chr,
                        VARDATA(email_list),
                        VARSIZE(email_list) - VARHDRSZ );
               email_list_chr[VARSIZE(email_list) - VARHDRSZ] = '\000';
               if( execl( SENDMAIL_PATH, 
                          SENDMAIL_PATH, 
                          email_list_chr, NULL ) == -1 )
                  {
                  syslog( LOG_ERR, "failed to exec SENDMAIL: %m" );
                  free( email_list_chr );
                  _exit( -1 );
                  }
               }
            else
               {
               syslog( LOG_ERR, "failed to malloc email_list_chr" );
               _exit( -1 );
               }
            break;
            }

         default:
            close( pipe_files[READ_PIPE] );
            write_msg( pipe_files[WRITE_PIPE],
                       msg_type,
                       pr_num,
                       project,
                       pr_title,
                       descr,
                       fix_descr,
                       status,
                       severity,
                       pr_type,
                       responsible,
                       submitter );
            close( pipe_files[WRITE_PIPE] );
            wait( &child_status );
            if( child_status != 0 )
               {
               syslog( LOG_ERR, "SENDMAIL returned: %d", child_status );
               }
         } /* end switch (fork) */
      } /* end if (pipe valid) */
   else
      {
      syslog( LOG_ERR, "pipe failed: %m" );
      }
}


/*-------------------------------------------------------------------------
 *-------------------------------------------------------------------------
 */
int4
send_new_pr_msg( const text *email_list,
                 const text *project,
                 const text *pr_title,
                 const text *descr,
                 const text *severity,
                 const text *problem_type,
                 const text *submitter )
{
   sendmsg( NEW_PR_MSG, email_list, INVALID_PR_NUM, project, pr_title, 
            descr, NULL, NULL, severity, problem_type, NULL, submitter );

   return 0;
}


/*-------------------------------------------------------------------------
 *-------------------------------------------------------------------------
 */
int4
send_assign_msg( const text *email_list,
                 const text *pr_title,
                 const text *descr,
                 const text *fix_descr,
                 const text *status,
                 const text *severity,
                 const text *responsible,
                 const text *submitter )
{
   sendmsg( ASSIGN_MSG, email_list, INVALID_PR_NUM, NULL, pr_title, descr, 
            fix_descr, status, severity, NULL, responsible, submitter );

   return 0;
}


/*-------------------------------------------------------------------------
 *-------------------------------------------------------------------------
 */
int4
send_update_msg( const text *email_list,
                 const text *pr_title,
                 const text *descr,
                 const text *fix_descr,
                 const text *status,
                 const text *severity,
                 const text *responsible,
                 const text *submitter )
{
   sendmsg( UPDATE_MSG, email_list, INVALID_PR_NUM, NULL, pr_title, descr, 
            fix_descr, status, severity, NULL, responsible, submitter );

   return 0;
}
